/*  Gpasman, a password manager
    Copyright (C) 1998, 1999 Olivier Sessink, olivier@lx.student.wau.nl

    Other code contributors:    
    Dave Rudder 
    Chris Halverson
    Matthew Palmer
    Guide Berning
    Jimmy Mason
	 website at http://www.student.wau.nl/~olivier/gpasman/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#define GPASMAN_NAME "Gpasman 1.3.0"

#include <gtk/gtk.h>
#include <string.h>
#include <stdlib.h>

#include "file.h"

#include "gpasman_icon.xpm"


/* this struct is used a couple of times, for changing the master password and adding 
or updating an entry */
typedef struct {
   GtkWidget *window;
   GtkWidget *entry1;
   GtkWidget *entry2;
   GtkWidget *entry3;
   GtkWidget *entry4;
   gint type;
} gtkentry_struct;


/* this is the struct for global variable current_set */
typedef struct {
   gchar *password;
   gint modified;
   gchar *filename;
} current_struct;


gint main (int argc, gchar * argv[]);
void get_main_menu( GtkWidget  *window, GtkWidget **menubar );
void make_interface(void);
void selection_made (GtkWidget * clist, gint row, gint column, GdkEventButton * event, gpointer data);
void unselection_made (GtkWidget * clist, gint row, gint column, GdkEventButton * event, gpointer data);
void click_made (GtkWidget * clist, gint column, gpointer data);
void toggle_show_passwords(gpointer callback_data,guint callback_action, GtkWidget *toggle);
void clist_button_press (GtkWidget *widget, GdkEvent *event, gpointer callback_data );
gchar *return_password();
void password_ok(GtkWidget *w, gpointer data);
void load_file ();
void save_file();
gint clist_get_entry(gint rownum, gchar *entry[4]);
void file_clear(GtkWidget *w, gpointer data);
gint modified_check(gchar *message, gchar *buttontext);
void modified_check_ok(GtkWidget *w, gpointer data);
void file_open(GtkWidget *w, gpointer data);
void file_save(GtkWidget *w, gpointer data);
void file_save_as(GtkWidget *w, gpointer data);
void entry_new(GtkWidget *w, gpointer data);
void entry_change(GtkWidget *w, gpointer data);
void entry_remove(GtkWidget *w, gpointer data);
void entry_dialog(gint type);
void window_close (GtkWidget *w, gtkentry_struct *data);
void entry_dialog_ok (GtkWidget *w, gtkentry_struct *data);
gchar *
 return_file();
void modal_fs_ok_clicked(GtkWidget * widget, gpointer window);
void close_modal_window(GtkWidget * widget, gpointer window);
void change_password (GtkWidget * button, gpointer data);
void check_password (GtkWidget * button, gtkentry_struct *data);
void statusbar_message(gpointer message, gint time);
gint statusbar_remove(gpointer message_id);
void gpasman_exit(GtkWidget * button, gtkentry_struct *data);
void gpasman_about(GtkWidget * button, gtkentry_struct *data);
void gpasman_help(GtkWidget * button, gtkentry_struct *data);
void info_dialog(gchar *message);



/* global variables */
GtkWidget *main_window, *clist, *statusbar, *clipb_entry;
current_struct current_set;
gint selectedrow, lastselectedrow, context_id;

gint main (int argc, gchar * argv[]) {

   gchar *filename;

   gtk_init (&argc, &argv);
   make_interface();
   filename = g_malloc(1024);
   filename = strcat(strcpy(filename,getenv("HOME")), "/.gpasman");
#ifdef DEBUG
   g_print("main, filename = %s\n", filename);
#endif
   current_set.filename = g_malloc(strlen(filename)+1);
   strcpy(current_set.filename, filename);
#ifdef DEBUG
   g_print("main, current_set.filename = %s\n", current_set.filename);
#endif
   g_free(filename);
   if (file_exists(current_set.filename) == 1) {
      load_file(current_set.filename);
      gtk_widget_show(main_window);
   } else {
      gtk_widget_show(main_window);
   }
   gtk_main();
   return 1;
}

static GtkItemFactoryEntry menu_items[] = {
       { "/_File",         NULL,         NULL, 0, "<Branch>" },
       { "/File/Clear",    NULL        , file_clear, 0, NULL },
       { "/File/_Open",    "<control>O", file_open, 0, NULL },
       { "/File/_Save",    "<control>S", file_save, 0, NULL },
       { "/File/Save _As", NULL,         file_save_as, 0, NULL },
       { "/File/sep1",     NULL,         NULL, 0, "<Separator>" },
       { "/File/Quit",     "<control>Q", gpasman_exit, 0, NULL },
       { "/_Edit",         NULL,         NULL, 0, "<Branch>" },
       { "/Edit/New entry",  NULL,       entry_new, 0, NULL },
       { "/Edit/Update entry",  NULL,    entry_change, 0, NULL },
       { "/Edit/Remove entry",  NULL,    entry_remove, 0, NULL },
       { "/Edit/sep1",     NULL,         NULL, 0, "<Separator>" },
       { "/Edit/Change password",  NULL, change_password, 0, NULL },
       { "/_View",         NULL,         NULL, 0, "<Branch>" },
       { "/View/Show passwords", NULL,   toggle_show_passwords, 0, "<ToggleItem>" },
       { "/_Help",         NULL,         NULL, 0, "<LastBranch>" },
       { "/_Help/Help",   NULL,         gpasman_help, 0, NULL },
       { "/_Help/About",   NULL,         gpasman_about, 0, NULL },
     };
     
void get_main_menu( GtkWidget  *window, GtkWidget **menubar ) {
   GtkItemFactory *item_factory;
   GtkAccelGroup *accel_group;
   gint nmenu_items = sizeof (menu_items) / sizeof (menu_items[0]);

   accel_group = gtk_accel_group_new ();
   item_factory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>", 
                                            accel_group);
   gtk_item_factory_create_items (item_factory, nmenu_items, menu_items, NULL);
   gtk_accel_group_attach (accel_group, GTK_OBJECT (window));
   if (menubar)
      *menubar = gtk_item_factory_get_widget (item_factory, "<main>");
}


void make_interface(void) {

   GdkPixmap *icon = NULL;
   GtkWidget *vbox, *hbox, *handle_box, *menubar, *scrolwin, *label;
   gchar *titles[4] = {"Host", "Username", "Password", "Comment"};
   GtkTooltips *tooltips;
   GtkCListColumn *column;
   
   main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
   gtk_widget_set_usize (GTK_WIDGET (main_window), 600, 300);
   gtk_window_set_title (GTK_WINDOW (main_window), GPASMAN_NAME);
   gtk_signal_connect (GTK_OBJECT (main_window),
			  "destroy",
			  GTK_SIGNAL_FUNC (gpasman_exit),
			  NULL);
   gtk_signal_connect (GTK_OBJECT (main_window),
			  "delete_event",
			  GTK_SIGNAL_FUNC (gpasman_exit),
			  NULL);
   gtk_window_set_wmclass(GTK_WINDOW(main_window), "Gpasman",
			   "gpasman");

   vbox = gtk_vbox_new (FALSE, 0);
   /* gtk_container_border_width (GTK_CONTAINER (vbox), 5); */
   gtk_container_add (GTK_CONTAINER (main_window), vbox);
   gtk_widget_show (vbox);
   
   handle_box = gtk_handle_box_new();
   gtk_box_pack_start (GTK_BOX(vbox), handle_box, FALSE, FALSE, 0);
   gtk_widget_show(handle_box);

   get_main_menu(main_window, &menubar);
   gtk_container_add (GTK_CONTAINER (handle_box), menubar);
   gtk_widget_show (menubar);
					    
   /* since GTK 1.2 one should use a scrolledwindow for a clist */
   scrolwin = gtk_scrolled_window_new (NULL, NULL);
   gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolwin),
				  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
   gtk_box_pack_start (GTK_BOX (vbox), scrolwin, TRUE, TRUE, 0);
   gtk_widget_show (scrolwin);

   tooltips = gtk_tooltips_new ();

   clist = gtk_clist_new_with_titles (4, titles);
   gtk_clist_column_titles_active (GTK_CLIST (clist));
   /* It isn't necessary to shadow the border, but it looks nice :) */
   gtk_clist_set_shadow_type (GTK_CLIST (clist), GTK_SHADOW_OUT);
   gtk_clist_set_column_width (GTK_CLIST (clist), 0, 175);
   gtk_clist_set_column_width (GTK_CLIST (clist), 1, 80);
   gtk_clist_set_column_width (GTK_CLIST (clist), 2, 80);
   gtk_clist_set_column_width (GTK_CLIST (clist), 3, 200);
   gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scrolwin), clist);
   gtk_clist_set_auto_sort (GTK_CLIST(clist), 1);
   gtk_clist_set_column_visibility(GTK_CLIST(clist), 2, 0);
   gtk_widget_show (clist);
   
   gtk_signal_connect(GTK_OBJECT (clist), "click_column",
		      GTK_SIGNAL_FUNC (click_made),
		      NULL);
   gtk_signal_connect(GTK_OBJECT (clist), "button_press_event",
		      GTK_SIGNAL_FUNC (clist_button_press),
		      NULL);
   gtk_signal_connect(GTK_OBJECT (clist), "select_row",
		      GTK_SIGNAL_FUNC (selection_made),
		      NULL);
   gtk_signal_connect(GTK_OBJECT (clist), "unselect_row",
		      GTK_SIGNAL_FUNC (unselection_made),
		      NULL);

   hbox = gtk_hbox_new(FALSE, 0);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
   gtk_widget_show(hbox);
   
   label = gtk_label_new(" Clipboard: ");
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_widget_show(label);
   
   clipb_entry = gtk_entry_new();
   gtk_box_pack_start (GTK_BOX (hbox), clipb_entry, FALSE, FALSE, 0);
   gtk_widget_show(clipb_entry);
   gtk_entry_set_editable(GTK_ENTRY(clipb_entry), 0);
   
   gtk_tooltips_set_tip (tooltips, clipb_entry, "Doubleclick an entry to copy the password", NULL);

   column = GTK_CLIST(clist)->column;
   gtk_tooltips_set_tip (tooltips, column->button, "Sort by host", NULL);
   column++;
   gtk_tooltips_set_tip (tooltips, column->button, "Sort by user", NULL);
   column++;
   gtk_tooltips_set_tip (tooltips, column->button, "Sort by password", NULL);
   column++;
   gtk_tooltips_set_tip (tooltips, column->button, "Sort by comment", NULL);
   
   statusbar = gtk_statusbar_new();
   gtk_box_pack_start (GTK_BOX (hbox), statusbar, TRUE, TRUE, 0);
   gtk_widget_show(statusbar);
   
   context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), "Gpasman: ");
   statusbar_message("Initialising..", 1000);
   
   gtk_widget_realize(main_window);
   icon = gdk_pixmap_create_from_xpm_d(main_window->window, NULL,
					NULL, (gchar **) gpasman_icon_xpm);
   gdk_window_set_icon(main_window->window, NULL, icon, NULL);
   gdk_window_set_icon_name(main_window->window, GPASMAN_NAME);
}

void selection_made (GtkWidget * clist, gint row, gint column, GdkEventButton * event, gpointer data) {
  selectedrow = lastselectedrow = row;
}

void unselection_made (GtkWidget * clist, gint row, gint column, GdkEventButton * event, gpointer data) {
  selectedrow = -1;
}

void click_made (GtkWidget * clist, gint column, gpointer data) {
   gtk_clist_set_sort_column (GTK_CLIST(clist), column);
   gtk_clist_sort(GTK_CLIST(clist));
}

void toggle_show_passwords(gpointer callback_data,guint callback_action, GtkWidget *toggle) {
   if ((GTK_CHECK_MENU_ITEM(toggle)->active)) {
      gtk_clist_set_column_visibility(GTK_CLIST(clist), 2, 1);
   } else {
      gtk_clist_set_column_visibility(GTK_CLIST(clist), 2, 0);
   }
}

void clist_button_press (GtkWidget *widget, GdkEvent *event, gpointer callback_data ) {

   gchar *text1=NULL;
   
   if (event->type == GDK_2BUTTON_PRESS) {
#ifdef DEBUG
      g_print("clist_button_press, 2 button press, lastselectedrow=%d\n", lastselectedrow);
#endif
     gtk_clist_get_text (GTK_CLIST (clist), lastselectedrow, 2, &text1);
#ifdef DEBUG
      g_print("clist_button_press, text1=%p\n", text1);
#endif
      if (text1) {
        gtk_entry_set_text(GTK_ENTRY(clipb_entry), text1);
        gtk_editable_select_region(GTK_EDITABLE(clipb_entry), 0, -1);
        gtk_editable_copy_clipboard(GTK_EDITABLE(clipb_entry));
      }
   }

}

gchar *return_password() {

   GtkWidget *vbox, *button, *label, *entry1, *window;
   gchar *tmp, *tmp2;

   window = gtk_window_new(GTK_WINDOW_DIALOG);
   gtk_container_border_width (GTK_CONTAINER (window), 10);
   gtk_window_position (GTK_WINDOW (window), GTK_WIN_POS_MOUSE);
   gtk_window_set_title (GTK_WINDOW (window), "Gpasman - enter master password");
   
   vbox = gtk_vbox_new(FALSE,5);
   gtk_container_add (GTK_CONTAINER (window), vbox);
   gtk_widget_show (vbox);

   label = gtk_label_new("Enter master password");
   gtk_box_pack_start (GTK_BOX (vbox),label , TRUE, TRUE, 0);
   gtk_widget_show(label);

   entry1 = gtk_entry_new_with_max_length(100);
   gtk_entry_set_visibility (GTK_ENTRY (entry1), 0);
   gtk_box_pack_start (GTK_BOX (vbox),entry1 , TRUE, TRUE, 0);
   gtk_widget_show(entry1);
   
   gtk_signal_connect(GTK_OBJECT (entry1), "activate",
		      GTK_SIGNAL_FUNC (password_ok),
		      NULL);
   
   button = gtk_button_new_with_label("Ok");
   gtk_box_pack_start (GTK_BOX (vbox),button , TRUE, TRUE, 0);
   gtk_signal_connect(GTK_OBJECT (button), "clicked",
		      GTK_SIGNAL_FUNC (password_ok),
		      NULL);
   gtk_widget_show(button);

   label = gtk_label_new(current_set.filename);
   gtk_box_pack_start (GTK_BOX (vbox),label , TRUE, TRUE, 0);
   gtk_widget_show(label);

   GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
   gtk_window_set_default(GTK_WINDOW(window), button);
   gtk_widget_show(window);
   gtk_window_set_focus(GTK_WINDOW(window), entry1);
      
   gtk_main();
   tmp = gtk_entry_get_text (GTK_ENTRY (entry1));

   if (strlen(tmp) < 4) {
      statusbar_message("Password has to be > 4 characters", 8000);
      gtk_widget_destroy(window);
      return NULL;
   }
#ifdef DEBUG
   g_print("return_password, tmp=%s\n", tmp);
#endif
   tmp2 = g_malloc(strlen(tmp)+1);
   strncpy(tmp2,tmp,strlen(tmp)+1);
   
   gtk_widget_destroy(window);
   
   return tmp2;
}

void password_ok(GtkWidget *w, gpointer data) {

   gtk_main_quit();
}

void load_file () {


   gchar *entry[4];
   gint retval, count;

   if (current_set.password) g_free(current_set.password);
   
   current_set.password = return_password();
   if (! current_set.password ) {
      return ;
   }
#ifdef DEBUG
   g_print("load_file, current_set.password=%s\n", current_set.password);
#endif

   retval = load_init(current_set.filename, current_set.password);
   switch (retval) {
      case 1:
         statusbar_message("Load initialised", 3000);
	 break;
      case 0:
         statusbar_message("Can't open file", 8000);
	 return;
      case -1:
         statusbar_message("File permissions bad, should be -rw------", 8000);
	 return;
      case -2:
         statusbar_message("File is a symlink", 8000);
	 return;
      case -3:
         statusbar_message("Can't get file status", 8000);
	 return;
      default:
         statusbar_message("Internal program error", 8000);
	 return;
   }
#ifdef DEBUG
   g_print("load_file, load_init retval=%d, entry=%p\n", retval, entry);
#endif
   retval = load_entry(entry);
   if (retval == 2) {
      statusbar_message("Last entry loaded", 3000);
   }
	 
#ifdef DEBUG
   g_print("load_file, load_entry retval=%d, entry=%p\n", retval, entry);
   g_print("load_file, entry[0]=%p, entry[1]=%p, entry[2]=%p, entry[3]=%p\n", entry[0], entry[1], entry[2], entry[3]);
#endif
   while (retval == 1) {
#ifdef DEBUGload
      g_print("load_file, entry[1]=%s\n", entry[1]);
#endif
      gtk_clist_append ((GtkCList *) clist, entry);
      for (count = 0; count < 4; count++) {
         g_free(entry[count]);
#ifdef DEBUG
      g_print("load_file, freeiing entry[%d]\n", count);
#endif
      }
      retval = load_entry(entry);
      if (retval == 2) {
         statusbar_message("Last entry loaded", 3000);
      }
#ifdef DEBUGload
      g_print("load_file, load_entry retval=%d\n", retval);
#endif
   }
   retval = load_finalize();
   current_set.modified = 0;
#ifdef DEBUGload
      g_print("load_file, load_finalize retval=%d\n", retval);
#endif
}


void save_file() {

   gint retval, retval2, rowcount;
   char *entry[4];

   if (! current_set.filename) file_save_as(NULL,NULL);
   if (! current_set.password) {
      current_set.password = return_password();
      if (! current_set.password ) {
         return ;
      }
   }
   
   retval = save_init(current_set.filename, current_set.password);
   switch (retval) {
      case 1:
         statusbar_message("Save initialised", 3000);
	 break;
      case 0:
         statusbar_message("Can't open/create file", 8000);
	 return;
      case -1:
         statusbar_message("File permissions bad, should be -rw------", 8000);
	 return;
      case -2:
         statusbar_message("File is a symlink", 8000);
	 return;
      case -3:
         statusbar_message("Can't get file status", 8000);
	 return;
      default:
         statusbar_message("Internal program error", 8000);
	 return;
   }
   
   rowcount = 0;
   retval2 = clist_get_entry(rowcount, entry);
   while (retval2 == 1) {
#ifdef DEBUG
      g_print("save_file, rowcount=%d\n", rowcount);
#endif
      if (save_entry(entry) == -1) {
         statusbar_message("Error writing file, contents not saved", 8000);
	 save_finalize();
	 return ;
      }
      /* error handling here! */
      rowcount++;
      retval2 = clist_get_entry(rowcount, entry);
   }
   if (save_finalize() == -1) {
      statusbar_message("Error writing file, contents not saved", 8000);
   } else {
      statusbar_message("Save finished", 3000);
      current_set.modified = 0;
   }

}

gint clist_get_entry(gint rownum, gchar *entry[4]) {

   gint count;
   gint tmp;
   gchar *text1;
   
   for (count = 0; count <4; count++) {
      tmp = gtk_clist_get_text (GTK_CLIST (clist), rownum, count, &text1);
#ifdef DEBUG
      g_print("clist_get_entry, tmp =%d, count=%d\n", tmp, count);
#endif
      if (tmp == 0) {
         return 0;
      }
      entry[count] = g_malloc(strlen(text1)+1);
      strcpy(entry[count], text1);
   }
   return 1;
}

void file_clear(GtkWidget *w, gpointer data) {
   if (modified_check("Contents are modified,\n clear contents?", " Clear ") ==1) {
      gtk_clist_clear (GTK_CLIST(clist));
      gtk_entry_set_text(GTK_ENTRY(clipb_entry), "");
      current_set.modified = 0;
      g_free(current_set.password);
      current_set.password = NULL;
      g_free(current_set.filename);
      current_set.filename = NULL;
   }
}


gint modified_check(gchar *message, gchar *buttontext) {

   GtkWidget *dialog, *open_button, *cancel_button, *label;
   gint check_to_return;
   
   if (current_set.modified ==1) {
      
      check_to_return = 0;
      dialog = gtk_dialog_new();
      gtk_window_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
      gtk_container_border_width(GTK_CONTAINER(dialog), 5);
      label = gtk_label_new (message);
      open_button = gtk_button_new_with_label(buttontext);
      cancel_button = gtk_button_new_with_label(" Cancel ");
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG(dialog)->action_area), open_button, FALSE, FALSE, 0);
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG(dialog)->action_area), cancel_button, FALSE, FALSE, 0);
      gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox),label);
      GTK_WIDGET_SET_FLAGS (cancel_button, GTK_CAN_DEFAULT);
      GTK_WIDGET_SET_FLAGS (open_button, GTK_CAN_DEFAULT);
      gtk_signal_connect(GTK_OBJECT (open_button), "clicked",
                               GTK_SIGNAL_FUNC (modified_check_ok), (gpointer)&check_to_return);
      gtk_signal_connect_object(GTK_OBJECT (cancel_button), "clicked",
                               GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
      gtk_widget_show_all(dialog);
      gtk_window_set_default(GTK_WINDOW(dialog), cancel_button);
      gtk_widget_realize(GTK_WIDGET(dialog));
      gtk_window_set_transient_for(GTK_WINDOW(dialog), GTK_WINDOW(main_window));
#ifdef DEBUG
      g_print("modified_check, check_to_return(%p) = %d\n", &check_to_return, check_to_return);
#endif
      gtk_main();
#ifdef DEBUG
      g_print("modified_check, check_to_return(%p) = %d\n", &check_to_return, check_to_return);
#endif
      gtk_widget_destroy(dialog);
      return check_to_return;
   } else {
      return 1;
   }
}

void modified_check_ok(GtkWidget *w, gpointer data) {

#ifdef DEBUG
   g_print("modified_check_ok, data(%p) = %d\n", data, *(gint *)data );
#endif

   *(gint *)data= 1;
   
#ifdef DEBUG
   g_print("modified_check_ok, data = %d\n", *(gint *)data);
#endif
   gtk_main_quit();
}

void file_open(GtkWidget *w, gpointer data) {

   gchar * tmpfilename;

   if (! modified_check("Contents are modified,\n open new file?", " Open ") == 1) {
      return;
   }
   tmpfilename = return_file();
   if (tmpfilename &&(file_exists(tmpfilename)==1)) {
      file_clear(w, data);
      current_set.filename = g_malloc(strlen(tmpfilename)+1);
      strcpy(current_set.filename, tmpfilename);
      load_file();
   } else {
      statusbar_message("Could not open file", 4000);
   }
   g_free(tmpfilename);
}

void file_save(GtkWidget *w, gpointer data) {
   if (current_set.filename) {
      save_file();
   } else {
     file_save_as(w,data);
   }
}

void file_save_as(GtkWidget *w, gpointer data) {

   gchar * tmpfilename;

   tmpfilename = return_file();
   if (tmpfilename) {
      g_free(current_set.filename);   
      current_set.filename = g_malloc(strlen(tmpfilename)+1);
      strcpy(current_set.filename, tmpfilename);
      save_file();
   }
   g_free(tmpfilename);
}

void entry_new(GtkWidget *w, gpointer data) {
   entry_dialog(1);
}

void entry_change(GtkWidget *w, gpointer data) {
   entry_dialog(2);
}

void entry_remove(GtkWidget *w, gpointer data) {
   gtk_clist_remove (GTK_CLIST(clist), selectedrow);
   current_set.modified = 1;
}

void entry_dialog(gint type) {
/*
 * type = 1 --> Add
 * type = 2 --> Change
 */

   gtkentry_struct *dialog;
   GtkWidget *frame, *label, *vbox, *hbox, *button3, *button2;
   gchar *text1;

   dialog = g_malloc(sizeof(gtkentry_struct));
   dialog->type = type;
   dialog->window = gtk_window_new(GTK_WINDOW_DIALOG);
   gtk_container_border_width (GTK_CONTAINER (dialog->window), 10);
   gtk_window_position (GTK_WINDOW (dialog->window), GTK_WIN_POS_MOUSE);
   
   if (type == 1) {
      frame = gtk_frame_new("Add a password");
      gtk_window_set_title (GTK_WINDOW (dialog->window), "Gpasman - add password");
   } else {
      gtk_window_set_title (GTK_WINDOW (dialog->window), "Gpasman - change password");
      frame = gtk_frame_new("Change password");
   }
   gtk_container_add(GTK_CONTAINER (dialog->window), frame);
   gtk_widget_show(frame);

   hbox = gtk_hbox_new (FALSE, 5);
   gtk_container_add (GTK_CONTAINER (frame), hbox);
   gtk_widget_show (hbox);
   vbox = gtk_vbox_new (FALSE, 5);
   gtk_box_pack_start (GTK_BOX (hbox), vbox, FALSE, FALSE, 5);
   gtk_widget_show (vbox);
   
   dialog->entry1 = gtk_entry_new_with_max_length (100);
   dialog->entry2 = gtk_entry_new_with_max_length (50);
   dialog->entry3 = gtk_entry_new_with_max_length (50);
   dialog->entry4 = gtk_entry_new_with_max_length (300);
   if ((dialog->type ==2 ) && (selectedrow != -1))
   {
      gtk_clist_get_text (GTK_CLIST (clist), selectedrow, 0, &text1);
      gtk_entry_set_text (GTK_ENTRY (dialog->entry1), text1);
      gtk_clist_get_text (GTK_CLIST (clist), selectedrow, 1, &text1);
      gtk_entry_set_text (GTK_ENTRY (dialog->entry2), text1);
      gtk_clist_get_text (GTK_CLIST (clist), selectedrow, 2, &text1);
      gtk_entry_set_text (GTK_ENTRY (dialog->entry3), text1);
      gtk_clist_get_text (GTK_CLIST (clist), selectedrow, 3, &text1);
      gtk_entry_set_text (GTK_ENTRY (dialog->entry4), text1);
   }
   label = gtk_label_new ("Server:");
   gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (vbox), dialog->entry1, FALSE, FALSE, 0);
   gtk_widget_show (dialog->entry1);
   label = gtk_label_new ("Username:");
   gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (vbox), dialog->entry2, FALSE, FALSE, 0);
   gtk_widget_show (dialog->entry2);
   label = gtk_label_new ("Password:");
   gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (vbox), dialog->entry3, FALSE, FALSE, 0);
   gtk_widget_show (dialog->entry3);
   label = gtk_label_new ("Comment:");
   gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
   gtk_widget_show (label);
   gtk_box_pack_start (GTK_BOX (vbox), dialog->entry4, FALSE, FALSE, 0);
   gtk_widget_show (dialog->entry4);

   hbox = gtk_hbox_new (TRUE, 0);
   gtk_container_add (GTK_CONTAINER (vbox), hbox);
   gtk_widget_show (hbox);

   button2 = gtk_button_new_with_label ("Cancel");
   if (type ==1) {
      button3 = gtk_button_new_with_label ("Add");
   } else {
      button3 = gtk_button_new_with_label ("Update");
   }
   gtk_signal_connect (GTK_OBJECT (button2), "clicked",
		      GTK_SIGNAL_FUNC (window_close),
		      dialog);
   gtk_signal_connect (GTK_OBJECT (button3), "clicked",
		      GTK_SIGNAL_FUNC (entry_dialog_ok),
		      (gpointer) dialog);
   gtk_box_pack_start (GTK_BOX (hbox), button2, TRUE, TRUE, 0);
   gtk_box_pack_start (GTK_BOX (hbox), button3, TRUE, TRUE, 0);

   gtk_widget_show (button2);
   gtk_widget_show (button3);

   GTK_WIDGET_SET_FLAGS (button2, GTK_CAN_DEFAULT);
   GTK_WIDGET_SET_FLAGS (button3, GTK_CAN_DEFAULT);
   gtk_window_set_default(GTK_WINDOW(dialog->window), button3);
   gtk_window_set_focus(GTK_WINDOW(dialog->window), dialog->entry1);
   gtk_widget_show (dialog->window);
   gtk_widget_realize(GTK_WIDGET(dialog->window));
   gtk_window_set_transient_for(GTK_WINDOW(dialog->window), GTK_WINDOW(main_window));
}


void window_close (GtkWidget *w, gtkentry_struct *data) {

   gtk_widget_destroy(GTK_WIDGET(data->window));
   if (data) {
      g_free(data);
   }
}

void entry_dialog_ok (GtkWidget *w, gtkentry_struct *data) {

   gchar *entry_text[4];
   gint count;

   entry_text[0] = gtk_entry_get_text (GTK_ENTRY (data->entry1));
   entry_text[1] = gtk_entry_get_text (GTK_ENTRY (data->entry2));
   entry_text[2] = gtk_entry_get_text (GTK_ENTRY (data->entry3));
   entry_text[3] = gtk_entry_get_text (GTK_ENTRY (data->entry4));
   for (count =0 ; count < 4 ; count++) {
     if (strlen(entry_text[count]) == 0) {
        strcpy(entry_text[count], " ");
     }
   }
   if (data->type == 1 ) {
      gtk_clist_append (GTK_CLIST(clist), entry_text);
   } else {
      for (count=0; count <4 ; count++) {
         gtk_clist_set_text (GTK_CLIST (clist), selectedrow, count, entry_text[count]);
      }
   }
   current_set.modified = 1;
   window_close(w, data);
}


/* just for the modal file selection, a glocal gchar pointer */
gchar *filename_to_return;

gchar *return_file()
{

    GtkWidget *fs;
    gchar *tmp;

    fs = gtk_file_selection_new("Select file");

    gtk_signal_connect(GTK_OBJECT(fs), "destroy", close_modal_window, fs);
    /* Connect the ok_button to file_ok_sel function */
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fs)->ok_button),
		       "clicked", (GtkSignalFunc) modal_fs_ok_clicked, fs);
    /* Connect the cancel_button to destroy the widget */
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fs)->cancel_button),
		       "clicked", (GtkSignalFunc) close_modal_window,
		       GTK_OBJECT(fs));
    gtk_widget_show(fs);
    gtk_grab_add(GTK_WIDGET(fs));
    gtk_main();
    tmp = filename_to_return;
    filename_to_return = NULL;
    return tmp;
}
void modal_fs_ok_clicked(GtkWidget * widget, gpointer window)
{
   if (gtk_file_selection_get_filename(GTK_FILE_SELECTION(window)) == NULL) {
      filename_to_return = NULL;
   } else {
   
     filename_to_return = g_malloc(strlen(gtk_file_selection_get_filename
					 (GTK_FILE_SELECTION(window))) + 1);
     strcpy(filename_to_return, gtk_file_selection_get_filename
	   (GTK_FILE_SELECTION(window)));
   }
   gtk_main_quit();
   gtk_grab_remove(GTK_WIDGET(window));
   gtk_signal_handlers_destroy(window);
   gtk_widget_destroy(GTK_WIDGET(window));
}

void close_modal_window(GtkWidget * widget, gpointer window)
{
    if (filename_to_return != NULL) {
	g_free(filename_to_return);
	filename_to_return = NULL;
    }
    gtk_main_quit();
    gtk_grab_remove(GTK_WIDGET(window));
    gtk_signal_handlers_destroy(window);
    gtk_widget_destroy(GTK_WIDGET(window));
}


void change_password (GtkWidget * button, gpointer data) {

   gtkentry_struct *dialog;
   GtkWidget *label, *button2, *button3, *vbox, *frame, *hbox;

   dialog = g_malloc(sizeof(gtkentry_struct));
   
   dialog->window = gtk_window_new (GTK_WINDOW_DIALOG);
   gtk_container_border_width (GTK_CONTAINER (dialog->window), 5);
   gtk_grab_add (dialog->window);
   gtk_window_position (GTK_WINDOW (dialog->window), GTK_WIN_POS_MOUSE);
   gtk_window_set_title (GTK_WINDOW (dialog->window), "Gpasman - change master password");
   gtk_signal_connect_object (GTK_OBJECT (dialog->window), "delete_event",
			     GTK_SIGNAL_FUNC (NULL),
			     (gpointer) dialog->window);
   frame = gtk_frame_new("Change master password");
   gtk_container_add (GTK_CONTAINER (dialog->window), frame);
   gtk_widget_show(frame);
   
   /* this hbox is only for better spacing inside the frame */
   hbox = gtk_hbox_new (FALSE, 5);
   gtk_container_add (GTK_CONTAINER (frame), hbox);
   gtk_widget_show (hbox);
   vbox = gtk_vbox_new (FALSE, 5);
   gtk_box_pack_start (GTK_BOX (hbox), vbox, FALSE, FALSE, 5);
   gtk_widget_show (vbox);


   label = gtk_label_new ("Enter old password");
   gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
   gtk_widget_show(label);
   dialog->entry1 = gtk_entry_new_with_max_length (25);
   gtk_box_pack_start (GTK_BOX (vbox), dialog->entry1, FALSE, FALSE, 1);
   gtk_widget_show(dialog->entry1);
   gtk_entry_set_visibility (GTK_ENTRY (dialog->entry1), 0);
   gtk_entry_set_text (GTK_ENTRY (dialog->entry1), "");
   label = gtk_label_new ("Enter new password");
   gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
   gtk_widget_show(label);
   dialog->entry2 = gtk_entry_new_with_max_length (25);
   gtk_box_pack_start (GTK_BOX (vbox), dialog->entry2, FALSE, FALSE, 1);
   gtk_widget_show(dialog->entry2);
   gtk_entry_set_visibility (GTK_ENTRY (dialog->entry2), 0);
   gtk_entry_set_text (GTK_ENTRY (dialog->entry2), "");
   label = gtk_label_new ("Confirm new password");
   gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
   gtk_widget_show(label);
   dialog->entry3 = gtk_entry_new_with_max_length (25);
   gtk_box_pack_start (GTK_BOX (vbox), dialog->entry3, FALSE, FALSE, 1);
   gtk_widget_show(dialog->entry3);
   gtk_entry_set_text (GTK_ENTRY (dialog->entry3), "");
   gtk_entry_set_visibility (GTK_ENTRY (dialog->entry3), 0);


   hbox = gtk_hbox_new (TRUE, 0);
   gtk_container_add (GTK_CONTAINER (vbox), hbox);
   gtk_widget_show (hbox);

   button2 = gtk_button_new_with_label ("Cancel");
   button3 = gtk_button_new_with_label ("Ok");

   gtk_signal_connect (GTK_OBJECT (button2), "clicked",
		      GTK_SIGNAL_FUNC (window_close),
		      (gpointer) dialog);
   gtk_signal_connect (GTK_OBJECT (button3), "clicked",
		      GTK_SIGNAL_FUNC (check_password),
		      (gpointer) dialog);
   gtk_box_pack_start (GTK_BOX (hbox), button2, TRUE, TRUE, 0);
   gtk_box_pack_start (GTK_BOX (hbox), button3, TRUE, TRUE, 0);

   GTK_WIDGET_SET_FLAGS (button2, GTK_CAN_DEFAULT);
   GTK_WIDGET_SET_FLAGS (button3, GTK_CAN_DEFAULT);

   gtk_widget_show (button2);
   gtk_widget_show (button3);
#ifdef DEBUG
   g_print ("change_password, window=%p\n", dialog->window);
#endif
   gtk_window_set_default(GTK_WINDOW(dialog->window), button3);
   gtk_window_set_focus(GTK_WINDOW(dialog->window), dialog->entry1);
   gtk_widget_show (dialog->window);
   gtk_widget_realize(GTK_WIDGET(dialog->window));
   gtk_window_set_transient_for(GTK_WINDOW(dialog->window), GTK_WINDOW(main_window));
}

void check_password (GtkWidget * button, gtkentry_struct *data)
{
  gchar tmppass[100];

#ifdef DEBUG
  g_print ("check_password, window=%p\n", data->window);
#endif
   if (current_set.password) {
      strncpy (tmppass, gtk_entry_get_text (GTK_ENTRY (data->entry1)), 99);
      if (strncmp (current_set.password, tmppass, 99) != 0)
      {
        window_close (button, data);
        statusbar_message("Current master password incorrect", 6000);
        return;
      }

      strncpy (tmppass, gtk_entry_get_text (GTK_ENTRY (data->entry2)), 99);
      if (strlen(tmppass) < 4) {
        window_close(button,data);
        statusbar_message("Master password has to be > 4 characters", 6000);
        return;
      }
      if (strncmp (tmppass, gtk_entry_get_text (GTK_ENTRY (data->entry3)), 99) != 0)
      {
        window_close (button, data);
        statusbar_message("New master passwords not identical", 6000);
        return;
      }
    
      g_free(current_set.password);
   } else {
      strncpy (tmppass, gtk_entry_get_text (GTK_ENTRY (data->entry2)), 99);
      if (strlen(tmppass) < 4) {
        window_close(button,data);
        statusbar_message("Master password has to be > 4 characters", 6000);
        return;
      }
   }
   current_set.password = g_malloc(strlen(tmppass)+1);
    
   strcpy (current_set.password, tmppass);
   statusbar_message("Master password changed", 3000);
   current_set.modified = 1;
   window_close (button, data);
}

void statusbar_message(gpointer message, gint time) {

   gint count;

   count = gtk_statusbar_push(GTK_STATUSBAR(statusbar), context_id, (gchar *)message);
   gtk_timeout_add(time, statusbar_remove, GINT_TO_POINTER(count));
}

gint statusbar_remove(gpointer message_id) {
   gtk_statusbar_remove(GTK_STATUSBAR(statusbar),context_id,GPOINTER_TO_INT(message_id));
   return 0;
}

void gpasman_exit(GtkWidget * button, gtkentry_struct *data) {

   if (modified_check("Contents are modified,\n are you sure you want to exit?", " Exit ") == 1) {
      gtk_main_quit();
   }
}

void gpasman_about(GtkWidget * button, gtkentry_struct *data) {
   info_dialog("Gpasman is written by Olivier Sessink\ngpasman@nl.linux.org\nhttp://gpasman.nl.linux.org/\n(C) 1998-1999\nbecause he forgot a password\n\nthe rc2 encryption is based on the rc2\nlibrary by Matthew Palmer");
}

void gpasman_help(GtkWidget * button, gtkentry_struct *data) {
   info_dialog("Gpasman - a password manager\n\nGpasman encrypts a lot of passwords using one master\n password. Every time you load or save a new file a\n master password is asked to (un)lock the contents. You\n can change this master password using \"change \n password\". The contents can be edited using \"new\n entry\", \"update entry\" and \"remove entry\". If you have\n a lot of entries, sort them by clicking on the titles\n of the list. Doubleclicking an entry puts the password\n on the clipboard so you can simply paste it into your\n password asking application using the middle mouse\n button. Errors messages are put only on the statusbar,\n no annoying dialogs. If present, the file ~/.gpasman will be\n opened at startup. The rest should be pretty self\n explaining stuff.");
}


void info_dialog(gchar *message) {
   GtkWidget *dialog, *label, *ok_button;

   dialog = gtk_dialog_new();
   gtk_window_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
   gtk_container_border_width(GTK_CONTAINER(dialog), 5);
   label = gtk_label_new (message);
   ok_button = gtk_button_new_with_label("  Ok  ");
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG(dialog)->action_area), ok_button, FALSE, FALSE, 0);
   gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox),label);

   GTK_WIDGET_SET_FLAGS (ok_button, GTK_CAN_DEFAULT);
   gtk_signal_connect_object(GTK_OBJECT (ok_button), "clicked",
                               GTK_SIGNAL_FUNC (gtk_widget_destroy), (gpointer )dialog);
   gtk_widget_show_all(dialog);
   gtk_window_set_default(GTK_WINDOW(dialog), ok_button);
   gtk_window_set_focus(GTK_WINDOW(dialog), ok_button);
}
